package edu.ucsc.facedetect;

import java.awt.Graphics;
import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.StringTokenizer;

import javax.imageio.ImageIO;
import javax.swing.JFrame;


public class Driver {

    
    public static void main(String[] args) {
        new Driver();
    }
    
    public Driver() {
//        trainThresholds();
//        trainClasifier();
//        testDetector();
        
//        drawBestFeatures();
    }
    
    public void testDetector() {
        Feature[] features = Feature.getAllFeatures();        
        importThresholds(features);        
        features = importClassifiers(features);
        
        File dir = new File("images/faces");
        for (File file : dir.listFiles()) {
            try {
                file = new File("images/IMG_1292.JPG");
                
                BufferedImage image = ImageIO.read(file);                
                IntegralImage ii = new IntegralImage(image);
                
                checkForFace(features, ii, file, image);
                break;
            }
            catch (Exception e) {
                e.printStackTrace();
            }
        }
    }
    
    public void checkForFace(Feature[] features, IntegralImage ii, File file, BufferedImage image) {
        
        int h = ii.height - 1;
        int w = ii.width - 1;
        
        double threshold = 0;
        double threshold2 = 0;
        double threshold5 = 0;
        double threshold20 = 0;
        double threshold50 = 0;
        double threshold200 = 0;
        double threshold1000 = 0;
        
        int f = 0;
        for (Feature feature : features) {
            threshold += feature.alpha;
            
            if (f == 2) threshold2 = 0.50*threshold;
            if (f == 5) threshold5 = 0.50*threshold;
            if (f == 20) threshold20 = 0.50*threshold;
            if (f == 50) threshold50 = 0.50*threshold;            
            if (f == 200) threshold200 = 0.50*threshold;            
            if (f == 1000) threshold1000 = 0.50*threshold;            
            f++;
        }
        
        threshold *= 0.5;
        
        int count = 0;
        final ArrayList<double[]> faces = new ArrayList<double[]>();
        
        long time = System.currentTimeMillis();
        
        double scale = 1;
        mainLoop:
        while (true) {            
            
            // see if features will be too big for the image
            if (scale*24 > w || scale*24 > h) {
                break;
            } 

            double value = 0;
            for (double y=0; y <= (h - (int)24*scale); y+=1.5*scale) {
                for (double x=0; x <= (w - (int)24*scale); x+=1.5*scale) {
                    value = 0;
                    
                    f = 0;
                    for (Feature feature : features) {
                        value += feature.alpha*getHypothesis(feature, feature.threshold, ii, (int)x, (int)y, scale);
                    
                        f++;
                        if (f == 2 && value < threshold2) break;
                        if (f == 5 && value < threshold5) break;
                        if (f == 20 && value < threshold20) break;
                        if (f == 50 && value < threshold50) break;
                        if (f == 200 && value < threshold200) break;
                        if (f == 1000 && value < threshold1000) break;
                    }
                    
                    if (value >= threshold) {
                        faces.add(new double[] {x, y, scale});                        
                        System.out.println(x + "  " + y + " " + scale + " " + value);
                    }
                    
                    count++;
                }                
            }
            
            scale*=1.25;
            
//            if (scale > 10) {
//                break;
//            }
        }
        
        System.out.println("Time: " + (System.currentTimeMillis() - time));
        
        new FaceDisplayer(faces, file);
        System.out.println("Faces: " + faces.size());
        System.out.println("count: " + count);
    }
    
    Feature[] videoFeatures = null;
    
    public ArrayList<double[]> videoCheckForFace(BufferedImage image) {
     
        if (videoFeatures == null) {
            Feature[] features = Feature.getAllFeatures();        
            importThresholds(features);        
            videoFeatures = importClassifiers(features);
            
            System.out.println("here");
            
            int f = 0;
            for (Feature feature : videoFeatures) {
                if (f <= 1) {
                    feature.threshold += 0.0;
                }
                else if (f <= 5) {
                    feature.threshold += 0.01;
                }
                else if (f <= 10) {
                    feature.threshold += 0.02;
                }
                else if (f <= 50) {
                    feature.threshold += 0.02;
                }
                else if (f <= 200) {
                    feature.threshold += 0.02;
                }
                else if (f <= 500) {
                    feature.threshold += 0.02;
                }
                else if (f <= 1000) {
                    feature.threshold += 0.03;
                }
                else if (f <= 2000) {
                    feature.threshold += 0.03;
                }
                
                f++;
            }
            
//            for (Feature feature : videoFeatures) {
//                feature.threshold += 0.002;
//            }
        }        

        return videoCheckForFace(videoFeatures, new IntegralImage(image));
    }


    public ArrayList<double[]> videoCheckForFace(Feature[] features, IntegralImage ii) {
        
        int h = ii.height - 1;
        int w = ii.width - 1;
        
        double threshold = 0;
        double threshold1 = 0;
        double threshold5 = 0;
        double threshold10 = 0;
        double threshold50 = 0;
        double threshold200 = 0;
        double threshold500 = 0;
        double threshold1000 = 0;

        double ratio = 0.75;
        int f = 0;
        for (Feature feature : features) {
            threshold += feature.alpha;
                        
            if (f == 1) threshold1 = 0.55*threshold;
            if (f == 5) threshold5 = 0.75*threshold;
            if (f == 10) threshold10 = 0.75*threshold;
            if (f == 50) threshold50 = 0.75*threshold;            
            if (f == 200) threshold200 = 0.8*threshold;            
            if (f == 500) threshold500 = 0.8*threshold;            
            if (f == 1000) threshold1000 = 0.75*threshold;            
            if (f == 2000) break;
            
//            if (f == 1000) threshold1000 = ratio*threshold;            
            f++;
        }
        
        threshold *= ratio;
        
        ArrayList<double[]> faces = new ArrayList<double[]>();        
        long time = System.currentTimeMillis();        
        
        int count = 0;
        
        // sub-windows: 70428
        double scale = 1;
        try { 
        while (true) {            
            
            // see if features will be too big for the image
            if (scale*24 > w || scale*24 > h) {
                break;
            } 

            double value = 0;
            for (double y=0; y <= (h - (int)24*scale); y+=scale) {
                for (double x=0; x <= (w - (int)24*scale); x+=scale) {
                    value = 0;
                    count++;
                    
                    f = 0;
                    for (Feature feature : features) {
                        value += feature.alpha*getHypothesis(feature, feature.threshold, ii, (int)x, (int)y, scale);
                        
                        f++;
                        if (f == 1 && value < threshold1) break;
                        if (f == 5 && value < threshold5) break;
                        if (f == 10 && value < threshold10) break;
                        if (f == 50 && value < threshold50) break;
                        if (f == 200 && value < threshold200) break;
                        if (f == 500 && value < threshold500) break;
                        if (f == 1000 && value < threshold1000) break;
                        
                        if (f == 2000) break;
                    }
                    
                    if (value >= threshold) {
                        faces.add(new double[] {x, y, scale});                        
//                        System.out.println(x + "  " + y + " " + scale + " " + value);
                    }
                }                
            }
            
            scale*=1.25;
        }
        }
        catch (Exception e) {}
        
        System.out.println(faces.size() + "/" + count);
        System.out.println("Time: " + (System.currentTimeMillis() - time));
        return faces;
    }
    
    public void trainClasifier() {
        Feature[] features = Feature.getAllFeatures();
                
        IntegralImage[] examples = ImageConverter.getInstance().getAllExamples();        
        importThresholds(features);        
        System.out.println("Data loaded beginning training");         
              
        features[255680].paint();
        
        // set the initial weights
        double[] weights = new double[1600];
        for (int i=0; i<weights.length; i++) {
            weights[i] = 1.0/1600.0;
        }

        // perform a trial
        int trials = 20;
        int[] classifiers = new int[trials];
        double alphas[] = new double[trials];
        
        for (int t=0; t<trials; t++) {
            
            // 1. normalize the weights
            double weightSum = 0;
            for (double weight : weights) {
                weightSum += weight;
            }
            
            for (int i=0; i<weights.length; i++) {
                weights[i] = weights[i]/weightSum;
            }
            
            // 2. choose the best feature
            int bestFeature = 0;
            double bestError = Double.MAX_VALUE;

            int f = 0;
            for (Feature feature : features) {
                double error = 0;
                
                int i = 0;
                for (IntegralImage example : examples) {
                    int h = getHypothesis(feature, feature.threshold, example);
                    int y = (i < 800) ? 1 : 0;
                    
                    if (y != h) {
                        error += weights[i];
                    }

                    i++;
                }
                
                if (error < bestError) {
                    bestError = error;
                    bestFeature = f;
                }
                
                f++;
            }
            
            // 3. choose the best feature (bestFeature)
            Feature feature = features[bestFeature];
            System.out.println(bestError);
            
            // 4. update the weights based on the selected feature
            double beta = bestError/(1 - bestError);            
            
            int i = 0;
            for (IntegralImage example : examples) {
                int h = getHypothesis(feature, feature.threshold, example);
                int y = (i < 800) ? 1 : 0;
                
                // decrease the weight if the prediction is correct
                if (y == h) {
                    weights[i] = weights[i]*beta;
                }

                i++;
            }
            
            // record the chosen feature and weight
            classifiers[t] = bestFeature;
            alphas[t] = -Math.log(beta);
            
            
            System.out.println("Best Feature: " + bestFeature);
            System.out.println("Alpha: " + alphas[t]);
            System.out.println("Finsihed iteration: " + t);
        }
        
        try {
            BufferedWriter writer = new BufferedWriter(new FileWriter("data/bestFeatures.dat"));
         
            for (int t=0; t<trials; t++) {
                writer.write(classifiers[t] + " " + alphas[t] + "\n");
            }
            
            writer.close();
        }
        catch (Exception e) {
            e.printStackTrace();
        }
            
    }

    public void trainThresholds() {
        Feature[] features = Feature.getAllFeatures();
        IntegralImage[] examples = ImageConverter.getInstance().getAllExamples();
        
        System.out.println("examples: " + examples.length);
        
        double[] thresholds = new double[features.length];
        int[] errors = new int[features.length];
        
        System.out.println("Data loaded, beginning training");
        long time = System.currentTimeMillis();
        
        int count = 0;
        for (Feature feature : features) {
            
            if (count%1000 == 0) {
                System.out.println(count);
            }

            int best = 1600;
            double bestThreshold = 1;            
            double threshold = 1;
            
            if (feature.rectangles > 2) {
                threshold = 2;
            }

            for (int t=0; t<=40; t++) {
                int error = 0;
                                
               int i = 0;
                for (IntegralImage example: examples) {                    
                    int h = getHypothesis(feature, threshold, example);
                  
                    if (i < 800) {
                       if (h == 0) error++;
                    }
                    else {
                       if (h == 1) error++;
                    }
                    
                    i++;
                }

                if (error < best) {
                    best = error;
                    bestThreshold = threshold;
                }

                threshold -= 0.05;
                if (feature.rectangles > 2) {
                    threshold -= 0.05;
                }
            }
            
            thresholds[count] = bestThreshold;
            errors[count] = best;

           count++;
        }        
     
        try {
            BufferedWriter writer = new BufferedWriter(new FileWriter("data/featureThresholds.dat"));
         
            for (int i=0; i<thresholds.length; i++) {                
                writer.write(thresholds[i] + " " + errors[i] + "\n");
            }
            
            writer.close();
        }
        catch (Exception e) {
            e.printStackTrace();
        }
                
        System.out.println("Training finished: " + (System.currentTimeMillis() - time));
    }
    

    
    public int getHypothesis(Feature feature, double threshold, IntegralImage example) {
        double diff = getDiff(feature, example)/255.0/(double)feature.width/(double)feature.height;
        
        return (diff < threshold) ? 1 : 0;
    }
    
    int x = 0;
    int y = 0;
    int w = 0;
    int h = 0;
    public double getDiff(Feature feature, IntegralImage example) {
        x = feature.x;
        y = feature.y;
        w = feature.width;
        h = feature.height;
        
        return getDiff(feature, example, x, y, w, h);
    }
            
    public final int getHypothesis(Feature feature, double threshold, IntegralImage example, int x, int y, double scale) {
        
        double diff = getDiff(feature, example, x, y, scale)/255.0;
        
        return (diff < threshold) ? 1 : 0;
    }

    int sx1;
    int ex1;
    int diffX1;
    int sy1;
    int ey1;
    int diffY1;

    int sx2;
    int ex2;
    int diffX2;
    int sy2;
    int ey2;
    int diffY2;

    int sx3;
    int ex3;
    double diffX3; 
    int sy3;
    int ey3;
    double diffY3; 
    
    public final double getDiff(Feature feature, IntegralImage example, int x, int y, double scale) {
        
        sx1 = (int)Math.floor(x + scale*feature.x);
        ex1 = (int)Math.floor(x + scale*(feature.x + feature.width));
        diffX1 = ex1 - sx1;
        
        sx2 = (int)Math.floor(x + scale*(feature.x + feature.width));
        ex2 = (int)Math.floor(x + scale*(feature.x + 2.0*feature.width));
        diffX2 = ex2 - sx2;
        
        sy1 = (int)Math.floor(y + scale*feature.y);
        ey1 = (int)Math.floor(y + scale*(feature.y + feature.height));
        diffY1 = ey1 - sy1;
        
        sy2 = (int)Math.floor(y + scale*(feature.y + feature.height));
        ey2 = (int)Math.floor(y + scale*(feature.y + 2.0*feature.height));
        diffY2 = ey2 - sy2;
        
        // 2 rectangle feature
        if (feature.rectangles == 2) {
            
            // vertically adjacent rectangles
            if (feature.vertical) {
                if (!feature.darkFirst) {
                    return example.getValue(sx1, sy2, ex1, ey2)/diffX1/diffY2 - example.getValue(sx1, sy1, ex1, ey1)/diffX1/diffY1;             
                }
                else {
                    return example.getValue(sx1, sy1, ex1, ey1)/diffX1/diffY1 - example.getValue(sx1, sy2, ex1, ey2)/diffX1/diffY2;             
                }                                 
            }
            // horizontally adjacent rectangles
            else {
                if (!feature.darkFirst) {
                    return example.getValue(sx2, sy1, ex2, ey1)/diffX2/diffY1 - example.getValue(sx1, sy1, ex1, ey1)/diffX1/diffY1;             
                }
                else {
                    return example.getValue(sx1, sy1, ex1, ey1)/diffX1/diffY1 - example.getValue(sx2, sy1, ex2, ey1)/diffX2/diffY1;             
                }                
            }
        }
        // 3 rectangle feature
        else if (feature.rectangles == 3) {
                        
            // vertically adjacent rectangles
            if (feature.vertical) {
                sy3 = (int)Math.floor(y + scale*(feature.y + 2*feature.height));
                ey3 = (int)Math.ceil(y + scale*(feature.y + 3*feature.height));
                diffY3 = ey3 - sy3;

                if (!feature.darkFirst) {
                    return example.getValue(sx1, sy2, ex1, ey2)/diffX1/diffY2 - example.getValue(sx1, sy1, ex1, ey1)/diffX1/diffY1 - example.getValue(sx1, sy3, ex1, ey3)/diffX1/diffY3;                   
                }
                else {
                    return example.getValue(sx1, sy1, ex1, ey1)/diffX1/diffY1 - example.getValue(sx1, sy2, ex1, ey2)/diffX1/diffY2 + example.getValue(sx1, sy3, ex1, ey3)/diffX1/diffY3;             
                }                                 
            }
            // horizontally adjacent rectangles
            else {
                sx3 = (int)Math.floor(x + scale*(feature.x + 2*feature.width));
                ex3 = (int)Math.ceil(x + scale*(feature.x + 3*feature.width));
                diffX3 = ex3 - sx3;

                if (!feature.darkFirst) {
                    return example.getValue(sx2, sy1, ex2, ey1)/diffX2/diffY1 - example.getValue(sx1, sy1, ex1, ey1)/diffX1/diffY1 - example.getValue(sx3, sy1, ex3, ey1)/diffX3/diffY1;
                }
                else {
                    return example.getValue(sx1, sy1, ex1, ey1)/diffX1/diffY1 - example.getValue(sx2, sy1, ex2, ey1)/diffX2/diffY1 + example.getValue(sx3, sy1, ex3, ey1)/diffX3/diffY1;
                }                
            }
        }
        // 4 rectangle feature
        else {
            if (!feature.darkFirst) {
                return example.getValue(sx2, sy1, ex2, ey1)/diffX2/diffY1 - example.getValue(sx1, sy1, ex1, ey1)/diffX1/diffY1             
                     - example.getValue(sx2, sy2, ex2, ey2)/diffX2/diffY2 + example.getValue(sx1, sy2, ex1, ey2)/diffX2/diffY2;
            }
            else {
                return example.getValue(sx1, sy1, ex1, ey1)/diffX1/diffY1 - example.getValue(sx2, sy1, ex2, ey1)/diffX2/diffY1             
                     - example.getValue(sx1, sy2, ex1, ey2)/diffX1/diffY2 + example.getValue(sx2, sy2, ex2, ey2)/diffX2/diffY2;
            }    
        }
    }
    
    public double getDiff(Feature feature, IntegralImage example, int x, int y, int w, int h) {
            
        // 2 rectangle feature
        if (feature.rectangles == 2) {
            
            // vertically adjacent rectangles
            if (feature.vertical) {
                if (!feature.darkFirst) {
                    return example.getValue(x, y + h, x + w, y + 2*h) - example.getValue(x, y, x + w, y + h);             
                }
                else {
                    return example.getValue(x, y, x + w, y + h) - example.getValue(x, y + h, x + w, y + 2*h);             
                }                                 
            }
            // horizontally adjacent rectangles
            else {
                if (!feature.darkFirst) {
                    return example.getValue(x + w, y, x + 2*w, y + h) - example.getValue(x, y, x + w, y + h);             
                }
                else {
                    return example.getValue(x, y, x + w, y + h) - example.getValue(x + w, y, x + 2*w, y + h);             
                }                
            }
        }
        // 3 rectangle feature
        else if (feature.rectangles == 3) {
            
            // vertically adjacent rectangles
            if (feature.vertical) {
                if (!feature.darkFirst) {
                    return example.getValue(x, y + h, x + w, y + 2*h) - example.getValue(x, y, x + w, y + h) - example.getValue(x, y + 2*h, x + w, y + 3*h);                   
                }
                else {
                    return example.getValue(x, y, x + w, y + h) - example.getValue(x, y + h, x + w, y + 2*h) + example.getValue(x, y + 2*h, x + w, y + 3*h);             
                }                                 
            }
            // horizontally adjacent rectangles
            else {
                if (!feature.darkFirst) {
                    return example.getValue(x + w, y, x + 2*w, y + h) - example.getValue(x, y, x + w, y + h) - example.getValue(x + 2*w, y, x + 3*w, y + h);
                }
                else {
                    return example.getValue(x, y, x + w, y + h) - example.getValue(x + w, y, x + 2*w, y + h) + example.getValue(x + 2*w, y, x + 3*w, y + h);             
                }                
            }
        }
        // 4 rectangle feature
        else {
            if (!feature.darkFirst) {
                return example.getValue(x + w, y, x + 2*w, y + h) - example.getValue(x, y, x + w, y + h)
                     - example.getValue(x + w, y + h, x + 2*w, y + 2*h) + example.getValue(x, y + h, x + w, y + 2*h);             
            }
            else {
                return example.getValue(x, y, x + w, y + h) - example.getValue(x + w, y, x + 2*w, y + h)
                     - example.getValue(x, y + h, x + w, y + 2*h) + example.getValue(x + w, y + h, x + 2*w, y + 2*h);             
            }    
        }
    }
    
    public Feature[] importClassifiers(Feature[] features) {
        try {
            BufferedReader reader = new BufferedReader(new FileReader("data/bestFeatures.dat"));
            
            Feature[] best = new Feature[5000];
            
            String line = reader.readLine();
            int i=0;
            while (line != null) {
                StringTokenizer tokenizer = new StringTokenizer(line);

                best[i] = features[Integer.parseInt(tokenizer.nextToken())];
                best[i].alpha = Double.parseDouble(tokenizer.nextToken());
                                               
                line = reader.readLine();
                i++;
                
                if (i == best.length) {
                    break;
                }
            }
            
            reader.close();
            return best;
        }
        catch (Exception e) {
            e.printStackTrace();
        }
        
        return null;
    }

    
    public void importThresholds(Feature[] features) {
        try {
            BufferedReader reader = new BufferedReader(new FileReader("data/featureThresholds.dat"));
            
            String line = reader.readLine();
            int i=0;
            while (line != null) {
                StringTokenizer tokenizer = new StringTokenizer(line);                
                features[i].threshold = Double.parseDouble(tokenizer.nextToken());
                line = reader.readLine();
                i++;
            }
            
            reader.close();
        }
        catch (Exception e) {
            e.printStackTrace();
        }
    }
    
    public void drawBestFeatures() {
        Feature[] features = Feature.getAllFeatures();        
        importThresholds(features);        
        features = importClassifiers(features);
        
        double[][] featureImage = new double[24][24];
        for (int i=0; i<24; i++) {
            for (int j=0; j<24; j++) {
                featureImage[i][j] = 0.5;
            }            
        }
        
        for (Feature feature : features) {
            getDiff1(feature, feature.x, feature.y, feature.width, feature.height, featureImage);
        }
     
        new FaceDisplayer(featureImage);
        System.out.println(features.length);
    }
    
    public double getDiff1(Feature feature, int x, int y, int w, int h, double[][] data) {
        
        // 2 rectangle feature
        if (feature.rectangles == 2) {
            
            // vertically adjacent rectangles
            if (feature.vertical) {
                if (!feature.darkFirst) {
                    return updateData(x, y + h, x + w, y + 2*h, data, true) - updateData(x, y, x + w, y + h, data, false);             
                }
                else {
                    return updateData(x, y, x + w, y + h, data, true) - updateData(x, y + h, x + w, y + 2*h, data, false);             
                }                                 
            }
            // horizontally adjacent rectangles
            else {
                if (!feature.darkFirst) {
                    return updateData(x + w, y, x + 2*w, y + h, data, true) - updateData(x, y, x + w, y + h, data, false);             
                }
                else {
                    return updateData(x, y, x + w, y + h, data, true) - updateData(x + w, y, x + 2*w, y + h, data, false);             
                }                
            }
        }
        // 3 rectangle feature
        else if (feature.rectangles == 3) {
            
            // vertically adjacent rectangles
            if (feature.vertical) {
                if (!feature.darkFirst) {
                    return updateData(x, y + h, x + w, y + 2*h, data, true) - updateData(x, y, x + w, y + h, data, false) - updateData(x, y + 2*h, x + w, y + 3*h, data, false);                   
                }
                else {
                    return updateData(x, y, x + w, y + h, data, true) - updateData(x, y + h, x + w, y + 2*h, data, false) + updateData(x, y + 2*h, x + w, y + 3*h, data, true);             
                }                                 
            }
            // horizontally adjacent rectangles
            else {
                if (!feature.darkFirst) {
                    return updateData(x + w, y, x + 2*w, y + h, data, true) - updateData(x, y, x + w, y + h, data, false) - updateData(x + 2*w, y, x + 3*w, y + h, data, false);
                }
                else {
                    return updateData(x, y, x + w, y + h, data, true) + updateData(x + 2*w, y, x + 3*w, y + h, data, true) - updateData(x + w, y, x + 2*w, y + h, data, false);             
                }                
            }
        }
        // 4 rectangle feature
        else {
            if (!feature.darkFirst) {
                return updateData(x + w, y, x + 2*w, y + h, data, true) + updateData(x, y + h, x + w, y + 2*h, data, true)
                             - updateData(x, y, x + w, y + h, data, false)  - updateData(x + w, y + h, x + 2*w, y + 2*h, data, false);             
            }
            else {
                return updateData(x, y, x + w, y + h, data, true)  + updateData(x + w, y + h, x + 2*w, y + 2*h, data, true) 
                             - updateData(x + w, y, x + 2*w, y + h, data, false) - updateData(x, y + h, x + w, y + 2*h, data, false);             
            }    
        }
    }
    
    public double updateData(int x1, int y1, int x2, int y2, double[][] data, boolean add) {
        
        for (int y=y1; y<=y2; y++) {            
            for (int x=x1; x<=x2; x++) {
                
                if (add) {
                    data[y][x] += 0.005;
                }
                else {
                    data[y][x] -= 0.005;
                }
            }
        }
        
        return 0;
    }
}
