package edu.ucsc.facedetect;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Image;
import java.io.File;
import java.util.ArrayList;

import javax.imageio.ImageIO;
import javax.swing.JFrame;

/**
 * Feature class.
 */
public class Feature {

    public int rectangles;
    public boolean vertical;
    public boolean darkFirst;
    public int x;
    public int y;
    public int width;
    public int height;
    public double threshold;
    public double alpha;
    
    public Feature(int rectangles, boolean vertical, boolean darkFirst, int x, int y, int width, int height) {
        this.rectangles = rectangles;
        this.vertical = vertical;
        this.darkFirst = darkFirst;
        this.width = width;
        this.height = height;
        this.x = x;
        this.y = y;
    }
    
    /**
     * Exhaustively generates all 2,3, and 4 rectangle features.
     */
    public static Feature[] getAllFeatures() {
        ArrayList<Feature> features = new ArrayList<Feature>();

        // vertical 2 rectangle feature
        for (int h=1; h<=12; h++) {
            for (int w=1; w<=24; w++) {
                for (int x=0; x<(24-w); x++) {
                    for (int y=0; y<(24-2*h); y++) {
                        features.add(new Feature(2, true, false, x, y, w, h));
                        features.add(new Feature(2, true, true, x, y, w, h));
                    }                                                
                }                            
            }            
        }

        // horizontal 2 rectangle feature
        for (int h=1; h<=24; h++) {
            for (int w=1; w<=12; w++) {
                for (int x=0; x<(24-2*w); x++) {
                    for (int y=0; y<(24-h); y++) {
                        features.add(new Feature(2, false, false, x, y, w, h));
                        features.add(new Feature(2, false, true, x, y, w, h));
                    }                                                
                }                            
            }            
        }

        // vertical 3 rectangle feature
        for (int h=1; h<=8; h++) {
            for (int w=1; w<=24; w++) {
                for (int x=0; x<(24-w); x++) {
                    for (int y=0; y<(24-3*h); y++) {
                        features.add(new Feature(3, true, false, x, y, w, h));
                        features.add(new Feature(3, true, true, x, y, w, h));
                    }                                                
                }                            
            }            
        }

        // horizontal 3 rectangle feature
        for (int h=1; h<=24; h++) {
            for (int w=1; w<=8; w++) {
                for (int x=0; x<(24-3*w); x++) {
                    for (int y=0; y<(24-h); y++) {
                        features.add(new Feature(3, false, false, x, y, w, h));
                        features.add(new Feature(3, false, true, x, y, w, h));
                    }                                                
                }                            
            }            
        }
        
        // 4 rectangle feature
        for (int h=1; h<=12; h++) {
            for (int w=1; w<=12; w++) {
                for (int x=0; x<(24-2*w); x++) {
                    for (int y=0; y<(24-2*h); y++) {
                        features.add(new Feature(4, false, false, x, y, w, h));
                        features.add(new Feature(4, false, true, x, y, w, h));
                    }                                                
                }                            
            }            
        }
        
        return features.toArray(new Feature[0]);
    }
    
    public static void main(String[] args) {
        System.out.println("Features: " + getAllFeatures().length);
    }
    
    public void paint() {
        Image image = null;
        try {
            image = ImageIO.read(new File("smallface.bmp"));
        }
        catch (Exception e) {
            e.printStackTrace();
        }
        
        final Image i = image;
        
        System.out.println(darkFirst);
        System.out.println(rectangles);
        System.out.println(vertical);
        
      JFrame frame = new JFrame() {
          public void paint(Graphics g) {
              super.paint(g);
              
              //g.drawRect(20, 50, 240, 240);    
              g.drawImage(i, 20, 50, null);
//              
//              if (!vertical && rectangles == 3) {
//                  g.setColor(Color.WHITE);
//                  g.fillRect(20 + 10*x, 50 + 10*y, width*10, height*10);
//                  g.setColor(Color.BLACK);
//                  g.fillRect(20 + 10*(x + width), 50 + 10*y, width*10, height*10);
//                  g.setColor(Color.WHITE);
//                  g.fillRect(20 + 10*(x + 2*width), 50 + 10*y, width*10, height*10);                  
//                  g.setColor(Color.BLACK);
//                  g.drawRect(20 + 10*x, 50 + 10*y, width*30, height*10);                  
//              }              
//              if (vertical && rectangles == 2) {
//                  g.setColor(Color.BLACK);
//                  g.fillRect(20 + 10*x, 50 + 10*y, width*10, height*10);
//                  g.setColor(Color.WHITE);
//                  g.fillRect(20 + 10*x, 50 + 10*(y+height), width*10, height*10);
//                  g.setColor(Color.BLACK);
//                  g.drawRect(20 + 10*x, 50 + 10*y, width*10, height*20);                  
//                  
//              }
//              if (!vertical && rectangles == 2) {
//                  g.setColor(Color.WHITE);
//                  g.fillRect(20 + 10*x, 50 + 10*y, width*10, height*10);
//                  g.setColor(Color.BLACK);
//                  g.fillRect(20 + 10*(x+width), 50 + 10*y, width*10, height*10);
//                  g.setColor(Color.BLACK);
//                  g.drawRect(20 + 10*x, 50 + 10*y, width*20, height*10);                  
//                  
//              }
//              if (rectangles == 4) {
//                  g.setColor(Color.WHITE);
//                  g.fillRect(20 + 10*x, 50 + 10*y, width*10, height*10);
//                  g.setColor(Color.BLACK);
//                  g.fillRect(20 + 10*(x + width), 50 + 10*y, width*10, height*10);
//                  g.setColor(Color.BLACK);
//                  g.fillRect(20 + 10*x, 50 + 10*(y+height), width*10, height*10);
//                  g.setColor(Color.WHITE);
//                  g.fillRect(20 + 10*(x + width), 50 + 10*(y+height), width*10, height*10);
//                  g.setColor(Color.BLACK);
//                  g.drawRect(20 + 10*x, 50 + 10*y, width*0, height*20);                  
//                  
//              }
          }
      };
      
      frame.setSize(300, 350);
      frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      frame.setVisible(true);
    }
}
