package edu.ucsc.facedetect;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.StringTokenizer;

import javax.imageio.ImageIO;
import javax.swing.JFrame;

public class ImageConverter {
    
    private static ArrayList<int[]> faceBounds = new ArrayList<int[]>();
    
    private static ImageConverter instance;
        
    public static ImageConverter getInstance() {
        if (instance == null) {
            instance = new ImageConverter();
        }
        
        return instance;
    }
    
    private ImageConverter() {
        loadFaceOffsets();
    }
    
    public IntegralImage[] getAllExamples() {
        ArrayList<IntegralImage> examples = new ArrayList<IntegralImage>();

        examples.addAll(getIntegralImages("images/positives"));
        examples.addAll(getIntegralImages("images/negatives"));
        
        return examples.toArray(new IntegralImage[0]);
    }
    
    public ArrayList<IntegralImage> getIntegralImages(String dirFile) {
        File dir = new File(dirFile);
        ArrayList<IntegralImage> examples = new ArrayList<IntegralImage>();
        
        for (File file : dir.listFiles()) {
            int[][] data = new int[24][24];
            int y = 0;
            
            try {            
                BufferedReader reader = new BufferedReader(new FileReader(file));
                String line = reader.readLine();
                
                while (line != null) {
                    StringTokenizer tokenizer = new StringTokenizer(line, " ");
                    
                    for (int x=0; x<24; x++) {
                        data[y][x] = Integer.parseInt(tokenizer.nextToken());
                    }
                    
                    y++;
                    line = reader.readLine();
                }
                
                examples.add(new IntegralImage(data));
                reader.close();
            }
            catch (Exception e) {
                e.printStackTrace();
            }            
        }
        
        return examples;
    }

    public static void main(String[] args) {
//        getInstance().generatePositiveDataSet();
        getInstance().generateNegativeDataSet();
    }
    
    public static void generateNegativeDataSet() {
        File dir = new File("images/temp");

        int index = 0;
        for (File file : dir.listFiles()) {            
            
            try {
                BufferedImage image = ImageIO.read(file);                
                int[][] data = new int[24][24];
                index++;
//                System.err.println(image);
//                System.err.println(file);
                
                int iw = image.getWidth();
                int ih = image.getHeight();
                
                // pick a random subset of the image (25 - 50)%
                int w = (int)((Math.random()*0.25 + 0.25)*Math.min(iw, ih));
                int h = w;
                int x = (int)(Math.random()*(iw - w));
                int y = (int)(Math.random()*(ih - h));

                double pixels[] = image.getData().getPixels(0, 0, 
                        image.getWidth(), image.getHeight(), 
                        new double[3*image.getWidth()*image.getHeight()]);
                
                int max = 0;
                int min = 255;
                for (int j=0; j<24; j++) {
                    for (int i=0; i<24; i++) {
                        int count = 0;
                        double value = 0;
                        
                        int sx = x + (int)Math.floor(((double)i)/24.0*(double)w);
                        int ex = x + (int)Math.ceil(((double)i + 1)/24.0*(double)w);
                        int sy = y + (int)Math.floor(((double)j)/24.0*(double)w);
                        int ey = y + (int)Math.ceil(((double)j + 1)/24.0*(double)w);
                        
                        for (int y1=sy; y1<ey; y1++) {
                            for (int x1=sx; x1<ex; x1++) {
                                
                                value += 0.333*pixels[3*(image.getWidth()*y1 + x1) + 0];
                                value += 0.333*pixels[3*(image.getWidth()*y1 + x1) + 1];
                                value += 0.333*pixels[3*(image.getWidth()*y1 + x1) + 2];
                                count += 1;
                            }                            
                        }

                        int p = (int)(value/(double)count);
                        data[j][i] = p;
                        
                        min = Math.min(min, p);
                        max = Math.max(max, p);
                    }                    
                }
                
                for (int j=0; j<24; j++) {
                    for (int i=0; i<24; i++) {
                        data[j][i] = (int)((data[j][i] - min)*255.0/(double)(max - min));
                    }
                }
                
                File out = new File("images/negatives/" + index + ".dat");
                BufferedWriter writer = new BufferedWriter(new FileWriter(out));
                
                for (int j=0; j<24; j++) {
                    for (int i=0; i<24; i++) {
                        writer.write(data[j][i] + " ");
                    }
                    
                    writer.write("\n");
                }
                
                writer.close();                                
            }
            catch (Exception e) {
                e.printStackTrace();
            }
        }
    }
    
    public static void generatePositiveDataSet() {
        File dir = new File("images/faces");

        for (File file : dir.listFiles()) {          
            
            try {
                final BufferedImage image = ImageIO.read(file);                
                int index = Integer.parseInt(file.getName().substring(6, 10));
                final int[] bounds = faceBounds.get(index - 1);
                final int[][] data = new int[24][24];
                
                System.out.println(index);
                
                // get the face bounds
                int x = bounds[0];
                int y = bounds[1];
                int w = bounds[2] - bounds[0];
                int h = bounds[3] - bounds[1];
                
                // make the area square
                y += (h - w)/2;
                h = w;

                final double pixels[] = image.getData().getPixels(0, 0, 
                        image.getWidth(), image.getHeight(), 
                        new double[3*image.getWidth()*image.getHeight()]);
                
                int max = 0;
                int min = 255;
                for (int j=0; j<24; j++) {
                    for (int i=0; i<24; i++) {
                        int count = 0;
                        double value = 0;
                        
                        int sx = x + (int)Math.floor(((double)i)/24.0*(double)w);
                        int ex = x + (int)Math.ceil(((double)i + 1)/24.0*(double)w);
                        int sy = y + (int)Math.floor(((double)j)/24.0*(double)w);
                        int ey = y + (int)Math.ceil(((double)j + 1)/24.0*(double)w);
                        
                        for (int y1=sy; y1<ey; y1++) {
                            for (int x1=sx; x1<ex; x1++) {
                                
                                value += 0.333*pixels[3*(image.getWidth()*y1 + x1) + 0];
                                value += 0.333*pixels[3*(image.getWidth()*y1 + x1) + 1];
                                value += 0.333*pixels[3*(image.getWidth()*y1 + x1) + 2];
                                count += 1;
                            }                            
                        }

                        int p = (int)(value/(double)count);
                        data[j][i] = p;
                        
                        min = Math.min(min, p);
                        max = Math.max(max, p);
                    }                    
                }
                
                for (int j=0; j<24; j++) {
                    for (int i=0; i<24; i++) {
                        data[j][i] = (int)((data[j][i] - min)*255.0/(double)(max - min));
                    }
                }
                
//                JFrame frame = new JFrame() {
//                    public void paint(Graphics g) {
//                        super.paint(g);
//                        
//                        for (int j=0; j<24; j++) {
//                            for (int i=0; i<24; i++) {
//                                int p = data[j][i];
//                                System.out.println(p);
//                                
//                                g.setColor(new Color(p, p, p));
//                                g.fillRect(50+ 10*i, 50+10*j, 10, 10);
//                                                                g.setColor(Color.BLACK);g.drawRect(50, 50, 240, 240);
//
//                            }
//                        }
//
//                    }
//                };
//                
//                frame.setSize(1000, 700);
//                frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
//                frame.setVisible(true);
                
                File out = new File("images/positives/m" + file.getName().substring(6, 10) + ".dat");
                BufferedWriter writer = new BufferedWriter(new FileWriter(out));
                
                for (int j=0; j<24; j++) {
                    for (int i=0; i<24; i++) {
                        writer.write(data[j][23 - i] + " ");
                    }
                    
                    writer.write("\n");
                }
                
                writer.close();
            }
            catch (Exception e) {
                e.printStackTrace();
            }
        }
    }
        
    public static IntegralImage getIntegralImage(int x, int y, int w, int h, double[] pixels, int imageWidth) {
            
        int[][] data = new int[24][24];
                
        int max = 0;
        int min = 255;
        for (int j=0; j<24; j++) {
            for (int i=0; i<24; i++) {
                int count = 0;
                double value = 0;
                
                int sx = x + (int)Math.floor(((double)i)/24.0*(double)w);
                int ex = x + (int)Math.ceil(((double)i + 1)/24.0*(double)w);
                int sy = y + (int)Math.floor(((double)j)/24.0*(double)w);
                int ey = y + (int)Math.ceil(((double)j + 1)/24.0*(double)w);
                
                try {
                
                for (int y1=sy; y1<ey; y1++) {
                    for (int x1=sx; x1<ex; x1++) {
                        
                        value += 0.333*pixels[3*(imageWidth*y1 + x1) + 0];
                        value += 0.333*pixels[3*(imageWidth*y1 + x1) + 1];
                        value += 0.333*pixels[3*(imageWidth*y1 + x1) + 2];
                        count += 1;
                    }                            
                }
                
                }
                catch(Exception e) {
                    System.err.println(x + " " + y + " " + w);
                    System.err.println(sx + " " + ex + " " + sy + " " + ey);
                    
                    e.printStackTrace();
                    System.exit(0);
                }

                int p = (int)(value/(double)count);
                data[j][i] = p;
                
                min = Math.min(min, p);
                max = Math.max(max, p);
            }                    
        }
                
        for (int j=0; j<24; j++) {
            for (int i=0; i<24; i++) {
                data[j][i] = (int)((data[j][i] - min)*255.0/(double)(max - min));
            }
        }

        return new IntegralImage(data);
    }
    
    private void loadFaceOffsets() {
        File file = new File("images/data/faceData.out");
        
        try {            
            BufferedReader reader = new BufferedReader(new FileReader(file));
            String line = reader.readLine();
            
            while (line != null) {
                StringTokenizer tokenizer = new StringTokenizer(line, " ");
                
                // [x_bot_left y_bot_left x_top_left y_top_left ... 
                //  x_top_right y_top_right x_bot_right y_bot_right]
                double left1 = Double.parseDouble(tokenizer.nextToken());                
                double bottom1 = Double.parseDouble(tokenizer.nextToken());
                double left2 = Double.parseDouble(tokenizer.nextToken());                
                double top1 = Double.parseDouble(tokenizer.nextToken());
                double right1 = Double.parseDouble(tokenizer.nextToken());                
                double top2 = Double.parseDouble(tokenizer.nextToken());
                double right2 = Double.parseDouble(tokenizer.nextToken());                
                double bottom2 = Double.parseDouble(tokenizer.nextToken());
         
                int l = (int)(left1/2.0 + left2/2.0 + 0.5);
                int r = (int)(right1/2.0 + right2/2.0 + 0.5);
                int t = (int)(top1/2.0 + top2/2.0 + 0.5);
                int b = (int)(bottom1/2.0 + bottom2/2.0 + 0.5);
                faceBounds.add(new int[] {l, t, r, b});
                
                line = reader.readLine();
            }
            
            reader.close();
        }
        catch(Exception e) {
            e.printStackTrace();
        }
    }    
}
