package edu.ucsc.facedetect;

import java.awt.image.BufferedImage;

public class IntegralImage {
 
    int width = 25;
    int height = 25;
    
    int[] ii;
    double[] iiSquared;
    
    public IntegralImage(int[][] grayScaleImageData) {
        ii = new int[width*height];
        iiSquared = new double[width*height];
        
        int[][] s = new int[height][width];
        double[][] sSquared = new double[height][width];
        
        for (int i=0; i<width; i++) {
            s[i][0] = 0;
            s[0][i] = 0;
            ii[width*i] = 0;
            ii[i] = 0;
            
            sSquared[i][0] = 0;
            sSquared[0][i] = 0;
            iiSquared[width*i] = 0;
            iiSquared[i] = 0;
        }        
        
        for (int y=1; y<width; y++) {
            for (int x=1; x<width; x++) {
                s[y][x] = s[y - 1][x] + grayScaleImageData[y - 1][x - 1];
                ii[y*width + x] = ii[y*width + x - 1] + s[y][x];                
                
                int p = grayScaleImageData[y - 1][x - 1];
                sSquared[y][x] = sSquared[y - 1][x] + (p/255.0*p/255.0);
                iiSquared[y*width + x] = iiSquared[y*width + x - 1] + sSquared[y][x];                
            }
        }
    }

    public IntegralImage(BufferedImage image) {
        width = image.getWidth() + 1;
        height = image.getHeight() + 1;
        
        ii = new int[width*height];
        iiSquared = new double[width*height];
        
        int[][] s = new int[height][width];
        double[][] sSquared = new double[height][width];
        
        for (int i=0; i<width; i++) {
            s[0][i] = 0;
            ii[i] = 0;
            
            sSquared[0][i] = 0;
            iiSquared[i] = 0;
        }        

        for (int i=0; i<height; i++) {
            s[i][0] = 0;
            ii[width*i] = 0;
            
            sSquared[i][0] = 0;
            iiSquared[width*i] = 0;
        }

        double pixels[] = image.getData().getPixels(0, 0, 
                image.getWidth(), image.getHeight(), 
                new double[3*image.getWidth()*image.getHeight()]);
        
        int min = 255;
        int max = 0;
        for (int y=1; y<height; y++) {
            for (int x=1; x<width; x++) {
                                
                double pixel = 0.33*pixels[3*((y-1)*image.getWidth() + (x - 1))] + 
                               0.33*pixels[3*((y-1)*image.getWidth() + (x - 1)) + 1] +
                               0.33*pixels[3*((y-1)*image.getWidth() + (x - 1)) + 2];

                min = Math.min(min, (int)pixel);
                max = Math.max(max, (int)pixel);
            }
        }
        
        for (int y=1; y<height; y++) {
            for (int x=1; x<width; x++) {
                                
                double pixel = 0.33*pixels[3*((y-1)*image.getWidth() + (x - 1))] + 
                               0.33*pixels[3*((y-1)*image.getWidth() + (x - 1)) + 1] +
                               0.33*pixels[3*((y-1)*image.getWidth() + (x - 1)) + 2];
                
                int p = (int)((pixel - min)*255.0/(double)(max - min));
                
                s[y][x] = s[y - 1][x] + p;
                ii[y*width + x] = ii[y*width + x - 1] + s[y][x];                
                
                sSquared[y][x] = sSquared[y - 1][x] + (p/255.0*p/255.0);
                iiSquared[y*width + x] = iiSquared[y*width + x - 1] + sSquared[y][x];                
            }
        }
    }

    public final double getValue(int x1, int y1, int x2, int y2) {
        return ii[y2*width + x2] + ii[y1*width + x1] - ii[y1*width + x2] - ii[y2*width + x1];
    }
}
